# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Entity resolver for the requests module.

Entity resolvers are considered part of the service-layer. The resolver
is registered in Invenio-Requests via the "invenio_requests.entity_resolvers"
entry point.
"""

from invenio_requests.resolvers.default import RecordResolver
from invenio_requests.resolvers.default.records import RecordPKProxy

from ..communities.records.api import Community
from ..communities.services.permissions import CommunityNeed


class CommunityPKProxy(RecordPKProxy):
    """Resolver proxy for a Record entity using the UUID."""

    def get_need(self):
        """Return the user need of the community's owner."""
        # TODO this may become difficult once there's multiple levels
        #      of membership (owner, manager, curator, ...)
        #      -> which needs should be generated? None, and let the
        #         user create the set of required needs from the
        #         resolved entity? or keep the 'owner' need?
        comid = str(self._parse_ref_dict_id(self._ref_dict))
        return CommunityNeed(comid)


class CommunityResolver(RecordResolver):
    """Community entity resolver.

    The entity resolver enables Invenio-Requests to understand communities as
    receiver and topic of a request.
    """

    type_id = 'community'
    """Type identifier for this resolver."""

    def __init__(self):
        """Initialize the default record resolver."""
        super().__init__(
            Community, type_key=self.type_id, proxy_cls=CommunityPKProxy)

    def _reference_entity(self, entity):
        """Create a reference dict for the given record."""
        return {self.type_key: str(entity.id)}



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utilities."""

from elasticsearch_dsl import Q
from invenio_cache import current_cache
from invenio_records_resources.services.errors import PermissionDeniedError

from .communities.services.permissions import CommunityNeed

# from datetime import timedelta
# from functools import partial

# from flask import current_app, render_template, request
# from flask_babelex import gettext as _
# from flask_mail import Message
# from invenio_accounts.models import User
# from invenio_base.utils import obj_or_import_string
# from invenio_mail.tasks import send_email
# from invenio_pidstore.resolver import Resolver
# from invenio_records.api import Record
# from invenio_records_rest.utils import LazyPIDValue
# from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit
# from werkzeug.routing import BaseConverter
# from werkzeug.utils import cached_property

# # from .members.api import CommunityMember
# # from .members.models import CommunityMemberRole, CommunityMemberStatus


# def format_url_template(url_template, absolute=True, **kwargs):
#     """Formats a URL template-like string."""
#     scheme, netloc, path, query, fragment = urlsplit(url_template)
#     scheme = scheme or request.scheme
#     netloc = netloc or request.host

#     # TODO: Find a more specific exception to raise here (or custom)
#     if absolute:
#         assert netloc
#     path = path.format(**kwargs)
#     query = urlencode({
#         k: [v.format(**kwargs) for v in vals]
#         for k, vals in parse_qs(query).items()
#     }, doseq=True)
#     return urlunsplit((scheme, netloc, path, query, fragment))


# def send_invitation_email(membership, recipients, community, token=b''):
#     """Send email notification for a member invitation or request."""
#     msg = Message(
#         _("Community membership invitation"),
#         sender=current_app.config.get('SUPPORT_EMAIL'),
#         recipients=recipients,
#     )
#     template = ("member_invitation_email.tpl" if membership.request.is_invite
#                 else "member_request_email.tpl")
#     msg.body = render_template(
#         "invenio_communities/email/{}".format(template),
#         community=community,
#         days=timedelta(
#             seconds=current_app.config[
#                 "COMMUNITIES_MEMBERSHIP_REQUESTS_CONFIRMLINK_EXPIRES_IN"]
#         ).days,
#         membership=membership,
#         invitation_link=format_url_template(
#             current_app.config.get(
#                 'COMMUNITIES_INVITATION_LINK_URL',
#                 '/communities/{com_id}/members/requests/{mem_id}?token={token}'
#             ),
#             mem_id=str(membership.id),
#             com_id=community["id"],
#             token=token.decode("utf-8")
#         ),
#         community_members_link=format_url_template(
#             '/communities/{com_id}/members?tab=pending',
#             com_id=community["id"]
#         )
#     )
#     send_email.delay(msg.__dict__)


# def send_inclusion_email(community_record, recipients):
#     """Send email notification for a record invitation or request."""
#     msg = Message(
#         _("Community record inclusion"),
#         sender=current_app.config.get('SUPPORT_EMAIL'),
#         recipients=recipients,
#     )
#     template = ("record_invitation_email.tpl" if
#                 community_record.request.is_invite
#                 else "record_request_email.tpl")
#     msg.body = render_template(
#         "invenio_communities/email/{}".format(template),
#         community_record=community_record,
#         record_invitation_link=format_url_template(
#             current_app.config.get(
#                 'COMMUNITIES_RECORD_INVITATION_LINK_URL',
#                 '/records/{recid}'
#             ),
#             recid=community_record.record.pid.pid_value
#         ),
#         # TODO: make overridable
#         community_curation_link=format_url_template(
#             '/communities/{com_id}/curate?q=',
#             com_id=community_record.community["id"]
#         )
#     )
#     send_email.delay(msg.__dict__)


# # TODO: Remove when
# # https://github.com/inveniosoftware/invenio-records-rest/pull/294 is merged
# class LazyPIDConverter(BaseConverter):
#     """Converter for PID values in the route mapping."""

#     def __init__(self, url_map, pid_type, getter=None, record_class=None,
#                  object_type='rec'):
#         """Initialize the converter."""
#         super(LazyPIDConverter, self).__init__(url_map)
#         self.pid_type = pid_type
#         self.getter = getter
#         self.record_class = record_class
#         self.object_type = object_type

#     @cached_property
#     def resolver(self):
#         """PID resolver."""
#         record_cls = obj_or_import_string(
#             self.record_class, default=Record)
#         getter = obj_or_import_string(
#             self.getter,
#             default=partial(record_cls.get_record, with_deleted=True))
#         return Resolver(
#             pid_type=self.pid_type,
#             object_type=self.object_type,
#             getter=getter
#         )

#     def to_python(self, value):
#         """Resolve PID value."""
#         return LazyPIDValue(self.resolver, value)


# comid_url_converter = (
#     'lazy_pid(comid,'
#     'record_class="invenio_communities:Community",'
#     'object_type="com")'
# )


# # def set_default_admin(community):
# #     """Set the Admin in a newly created community."""
# #     user = User.query.get(community['created_by'])
# #     CommunityMember.create(
# #         community=community,
# #         role=CommunityMemberRole.ADMIN,
# #         user=user,
# #         status=CommunityMemberStatus.ACCEPTED
# #     )


def search_communities(service, identity):
    """Search for communities owned by the given identity.

    We cannot use high-level service functions here because the given
    identity may not be fully initialized yet, and thus fail permission
    checks.
    """
    # TODO this needs to be revisited once memberships are in!
    search_result = service._search(
        'search',
        identity,
        params={},
        es_preference=None,
        extra_filter=Q(
            "term",
            **{"access.owned_by.user": identity.id}
        ),
        permission_action='read',
    ).execute()

    return search_result


def load_community_needs(identity, service):
    """Add community-related needs to the freshly loaded identity.

    Note that this function is intended to be called as handler for the
    identity-loaded signal, where we don't have control over the handler
    execution order.
    Thus, the given identity may not be fully initialized and still missing
    some needs (e.g. 'authenticated_user'), so we cannot rely on high-level
    service functions because permission checks may fail.
    """
    if identity.id is None:
        # no user is logged in
        return

    # Cache keys
    #
    # The cache of communities must be invalidated on:
    # 1) on creation of a community (likely this one is modelled via membership
    #    in the future).
    # 2) add/remove/change of membership
    #
    # We construct the cache key for each membership entity (e.g. user,
    # role, system role). This way, once a membership is added/removed/updated
    # we can cache the list of associated communities for this entity.
    # Once a user logs in, we get the cache for each of the membership
    # entities and combine it into a single list.

    # Currently, only users are supported (no roles or system roles)
    cache_key = identity_cache_key(identity)
    communities = current_cache.get(cache_key)
    if communities is None:
        try:
            communities = []
            for c in search_communities(service, identity):
                communities.append(str(c.uuid))
            current_cache.set(cache_key, communities, timeout=24*3600)
        except PermissionDeniedError:
            communities = []


    # Add community needs to identity
    for c_id in communities:
        identity.provides.add(CommunityNeed(c_id))


def on_membership_change(identity=None):
    """Handler called when a membership is changed."""
    if identity is not None:
        current_cache.delete(identity_cache_key(identity))

def identity_cache_key(identity):
    """Make the cache key for storing the communities for a user."""
    return f"user-communities:{identity.id}"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Communities Service API config."""

from flask_babelex import lazy_gettext as _
from invenio_records_resources.services import FileServiceConfig
from invenio_records_resources.services.files.links import FileLink
from invenio_records_resources.services.records.components import \
    DataComponent, MetadataComponent
from invenio_records_resources.services.records.config import \
    RecordServiceConfig
from invenio_records_resources.services.records.config import \
    SearchOptions as SearchOptionsBase
from invenio_records_resources.services.records.facets import TermsFacet
from invenio_records_resources.services.records.links import RecordLink, \
    pagination_links

from invenio_communities.communities.records.api import Community

from ..schema import CommunitySchema
from .components import CommunityAccessComponent, PIDComponent
from .permissions import CommunityPermissionPolicy


class SearchOptions(SearchOptionsBase):
    """Search options."""

    facets = {
        'type': TermsFacet(
            field='metadata.type',
            label=_('Type'),
            value_labels={
                'organization': _('Organization'),
                'event': _('Event'),
                'topic': _('Topic'),
                'project': _('Project'),
            }
        ),
        'domain': TermsFacet(field='metadata.domains'),

    }


class CommunityServiceConfig(RecordServiceConfig):
    """Communities service configuration."""

    # Common configuration
    permission_policy_cls = CommunityPermissionPolicy

    # Record specific configuration
    record_cls = Community

    # Search configuration
    search = SearchOptions

    # Service schema
    schema = CommunitySchema

    links_item = {
        "self": RecordLink("{+api}/communities/{id}"),
        "self_html": RecordLink("{+ui}/communities/{id}"),
        "settings_html": RecordLink("{+ui}/communities/{id}/settings"),
        "logo": RecordLink("{+api}/communities/{id}/logo"),
        "rename": RecordLink("{+api}/communities/{id}/rename")
    }

    links_search = pagination_links("{+api}/communities{?args*}")
    links_user_search = pagination_links("{+api}/user/communities{?args*}")

    # Service components
    components = [
        MetadataComponent,
        PIDComponent,
        CommunityAccessComponent,
    ]


class CommunityFileServiceConfig(FileServiceConfig):
    """Configuration for community files."""

    record_cls = Community
    permission_policy_cls = CommunityPermissionPolicy

    file_links_item = {
        "self": FileLink("{+api}/communities/{id}/logo"),
    }



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Communities Service API."""

from elasticsearch_dsl import Q
from invenio_db import db
from invenio_records_resources.services.base import LinksTemplate
from invenio_records_resources.services.records import RecordService
from invenio_records_resources.services.uow import RecordCommitOp, unit_of_work
from marshmallow.exceptions import ValidationError


class CommunityService(RecordService):
    """community Service."""

    def __init__(self, config, files_service=None):
        """Constructor for CommunityService."""
        super().__init__(config)
        self._files = files_service

    @property
    def files(self):
        """Community files service."""
        return self._files

    def search_user_communities(
            self, identity, params=None, es_preference=None, **kwargs):
        """Search for records matching the querystring."""
        self.require_permission(identity, 'search_user_communities')

        # Prepare and execute the search
        params = params or {}
        search_result = self._search(
            'search',
            identity,
            params,
            es_preference,
            extra_filter=Q(
                "term",
                **{"access.owned_by.user": identity.id}
            ),
            permission_action='read',
            **kwargs).execute()

        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=LinksTemplate(self.config.links_user_search, context={
                "args": params
            }),
            links_item_tpl=self.links_item_tpl,
        )

    @unit_of_work()
    def rename(self, identity, id_, data, revision_id=None, raise_errors=True,
               uow=None):
        """Rename a community."""
        record = self.record_cls.pid.resolve(id_)

        self.check_revision_id(record, revision_id)

        # Permissions
        self.require_permission(identity, "rename", record=record)

        if 'id' not in data:
            raise ValidationError(
                'Missing data for required field.',
                field_name='id',
            )

        data, errors = self.schema.load(
            data,
            context={"identity": identity},
            raise_errors=raise_errors,     # if False, flow is continued with
            schema_args={'partial': True}  # data only containing valid data,
                                           # but errors are reported
        )                                  # (as warnings)

        # Run components
        self.run_components(
            'rename', identity, data=data, record=record, uow=uow)

        uow.register(RecordCommitOp(record, indexer=self.indexer))

        return self.result_item(
            self,
            identity,
            record,
            links_tpl=self.links_item_tpl,
        )

    def read_logo(self, identity, id_):
        """Read the community's logo."""
        record = self.record_cls.pid.resolve(id_)
        self.require_permission(identity, 'read', record=record)
        logo_file = record.files.get('logo')
        if logo_file is None:
            raise FileNotFoundError()
        return self.files.file_result_item(
            self.files,
            identity,
            logo_file,
            record,
            links_tpl=self.files.file_links_item_tpl(id_),
        )

    @unit_of_work()
    def update_logo(self, identity, id_, stream, content_length=None,
                    uow=None):
        """Update the community's logo."""
        record = self.record_cls.pid.resolve(id_)
        self.require_permission(identity, 'update', record=record)

        record.files['logo'] = stream
        uow.register(RecordCommitOp(record))

        return self.files.file_result_item(
            self.files,
            identity,
            record.files['logo'],
            record,
            links_tpl=self.files.file_links_item_tpl(id_),
        )

    @unit_of_work()
    def delete_logo(self, identity, id_, uow=None):
        """Delete the community's logo."""
        record = self.record_cls.pid.resolve(id_)
        deleted_file = record.files.pop('logo', None)
        if deleted_file is None:
            raise FileNotFoundError()

        uow.register(RecordCommitOp(record))

        return self.files.file_result_item(
            self.files,
            identity,
            deleted_file,
            record,
            links_tpl=self.files.file_links_item_tpl(id_),
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

from invenio_db import db
from invenio_requests.records.models import RequestMetadata
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy_utils.types import UUIDType

from ...communities.records.models import CommunityMetadata


class CommunityRelationMixin:
    """Model mixin to define a relationship between a communities and records.

    Usage:

    .. code-block:: python

        class CommunityRecordM2M(db.Model, CommunityRelationMixin):
            __record_model__ = MyParentRecord
    """

    __record_model__ = None
    __request_model__ = None

    @declared_attr
    def community_id(cls):
        """Foreign key to the related communithy."""
        return db.Column(
            UUIDType,
            db.ForeignKey(CommunityMetadata.id, ondelete="CASCADE"),
            primary_key=True,
        )

    @declared_attr
    def record_id(cls):
        """Foreign key to the related record."""
        return db.Column(
            UUIDType,
            db.ForeignKey(cls.__record_model__.id, ondelete="CASCADE"),
            primary_key=True,
        )

    @declared_attr
    def request_id(cls):
        """Foreign key to a related request."""
        return db.Column(
            UUIDType,
            db.ForeignKey(RequestMetadata.id, ondelete="SET NULL"),
            nullable=True,
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community database models."""

from invenio_db import db
from invenio_files_rest.models import Bucket
from invenio_records.models import RecordMetadataBase
from invenio_records_resources.records import FileRecordModelMixin
from sqlalchemy_utils.types import UUIDType


class CommunityMetadata(db.Model, RecordMetadataBase):
    """Represent a community."""

    __tablename__ = 'communities_metadata'

    # Enables SQLAlchemy-Continuum versioning
    __versioned__ = {}

    bucket_id = db.Column(UUIDType, db.ForeignKey(Bucket.id))
    bucket = db.relationship(Bucket)


class CommunityFileMetadata(db.Model, RecordMetadataBase, FileRecordModelMixin):
    """File associated with a community."""

    __record_model_cls__ = CommunityMetadata

    __tablename__ = 'communities_files'



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community views."""

from .api import create_communities_api_blueprint
from .ui import create_ui_blueprint

__all__ = (
    'create_communities_api_blueprint',
    'create_ui_blueprint'
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Integration with the Invenio-Requests module."""

from .resolver import CommunityResolver

__all__ = (
    'CommunityResolver',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test alembic recipes for Invenio-RDM-Records."""

import pytest
from invenio_db.utils import drop_alembic_version_table
# Invenio-Vocabularies define an Alembic migration that include subjects and
# affiliations tables. The SQLAlchemy models are however not registered by
# default. This means that when alembic creates the database vs when SQLAlchemy
# creates the database there's a discrepancy. Importing the subjects model
# here, ensures the model get's registered, and we avoid the discrepancy.
from invenio_vocabularies.contrib.subjects import models


def test_alembic(base_app, database):
    """Test alembic recipes."""
    db = database
    ext = base_app.extensions['invenio-db']

    if db.engine.name == 'sqlite':
        raise pytest.skip('Upgrades are not supported on SQLite.')

    # Check that this package's SQLAlchemy models have been properly registered
    tables = [x.name for x in db.get_tables_for_bind()]
    assert 'communities_metadata' in tables
    assert 'communities_files' in tables
    assert 'communities_metadata_version' in tables

    # Check that Alembic agrees that there's no further tables to create.
    assert not ext.alembic.compare_metadata()

    # Hack to not having to deal with mock_metadata depending on which order
    # tests are run in.
    cmp_len = 2 if 'mock_metadata' in tables else 0

    # Drop everything and recreate tables all with Alembic
    db.drop_all()
    drop_alembic_version_table()
    ext.alembic.upgrade()
    assert len(ext.alembic.compare_metadata()) == cmp_len

    # Try to upgrade and downgrade
    ext.alembic.stamp()
    ext.alembic.downgrade(target='96e796392533')
    ext.alembic.upgrade()
    assert len(ext.alembic.compare_metadata()) == cmp_len

    drop_alembic_version_table()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tests for the CLI."""

from faker import Faker

from invenio_communities.fixtures.demo import create_fake_community
from invenio_communities.fixtures.tasks import create_demo_community


def test_fake_demo_community_creation(app, db, location, es_clear):
    """Assert that demo community creation works without failing."""
    faker = Faker()
    create_demo_community(create_fake_community(faker))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
# Copyright (C) 2021 Graz University of Technology.
# Copyright (C) 2021 TU Wien.

#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community permissions."""

import operator
from functools import partial, reduce
from itertools import chain

from elasticsearch_dsl.query import Q
from flask_principal import Need, UserNeed
from invenio_access.permissions import any_user
from invenio_records_permissions.generators import AnyUser, \
    AuthenticatedUser, Generator, SystemProcess
from invenio_records_permissions.policies import BasePermissionPolicy

# TODO this is temporary and should be revisited when membership is a thing
CommunityNeed = partial(Need, "community")


# TODO: Move class to Invenio-Records-Permissions and make more reusable ()
class IfRestrictedBase(Generator):
    """IfRestricted generator.

    IfRestricted(
        'record', 'restricted', 'public'
        then_=[RecordPermissionLevel('view')],
        else_=[ActionNeed(superuser-access)],
    )
    """

    def __init__(self, field_getter,
                 field_name, then_value, else_value, then_, else_):
        """Constructor."""
        self.field_getter = field_getter
        self.field_name = field_name
        self.then_value = then_value
        self.else_value = else_value
        self.then_ = then_
        self.else_ = else_

    def generators(self, record):
        """Get the "then" or "else" generators."""
        if record is None:
            # TODO - when permissions on links are checked, the record is not
            # passes properly, causing below ``record.access`` to fail.
            return self.else_
        # TODO: Next statement should be made more reusable
        is_then_value = self.field_getter(record) or self.then_value
        return self.then_ if is_then_value == self.then_value else self.else_

    def needs(self, record=None, **kwargs):
        """Set of Needs granting permission."""
        needs = [
            g.needs(record=record, **kwargs) for g in self.generators(record)]
        return set(chain.from_iterable(needs))

    def excludes(self, record=None, **kwargs):
        """Set of Needs denying permission."""
        needs = [
            g.excludes(record=record, **kwargs) for g in self.generators(
                record)]
        return set(chain.from_iterable(needs))

    def make_query(self, generators, **kwargs):
        """Make a query for one set of generators."""
        queries = [g.query_filter(**kwargs) for g in generators]
        queries = [q for q in queries if q]
        return reduce(operator.or_, queries) if queries else None

    def query_filter(self, **kwargs):
        """Filters for current identity as super user."""
        # TODO: Next two statement should be made more reusable
        q_then = Q("match", **{self.field_name: self.then_value})
        q_else = Q("match", **{self.field_name: self.else_value})
        then_query = self.make_query(self.then_, **kwargs)
        else_query = self.make_query(self.else_, **kwargs)

        if then_query and else_query:
            return (q_then & then_query) | (q_else & else_query)
        elif then_query:
            return (q_then & then_query) | q_else
        elif else_query:
            return q_else & else_query
        else:
            return q_else


# TODO: Remove class once IfRestrictedBase has been moved.
class IfRestricted(IfRestrictedBase):
    """IfRestricted."""

    def __init__(self, field, then_, else_):
        """Initialize."""
        super().__init__(
            lambda r: getattr(r.access, field, None),
            f"access.{field}",
            "restricted",
            "public",
            then_,
            else_,
        )


class IfPolicyClosed(IfRestrictedBase):
    """If policy is closed."""

    def __init__(self, field, then_, else_):
        """Initialize."""
        super().__init__(
            lambda r: getattr(r.access, field, None),
            f"access.{field}",
            "open",
            "closed",
            then_,
            else_,
        )


class CommunityOwners(Generator):
    """Allows community owners."""

    def needs(self, record=None, **kwargs):
        """Enabling Needs."""
        if record:
            return [
                UserNeed(owner.owner_id)
                for owner in record.access.owned_by
                ]
        else:
            return [any_user]

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as owner."""
        users = [n.value for n in identity.provides if n.method == "id"]
        return Q(
            "terms",
            **{"access.owned_by.user": users}
            )


class CommunityPermissionPolicy(BasePermissionPolicy):
    """Permissions for Community CRUD operations."""

    can_create = [AuthenticatedUser(), SystemProcess()]

    can_read = [
        IfRestricted(
            'visibility',
            then_=[CommunityOwners()],
            else_=[AnyUser()]),
        ]

    can_update = [CommunityOwners(), SystemProcess()]

    can_delete = [CommunityOwners(), SystemProcess()]

    can_search = [AnyUser(), SystemProcess()]

    can_search_user_communities = [AuthenticatedUser(), SystemProcess()]

    can_rename = [CommunityOwners(), SystemProcess()]

    can_submit_record = [
        IfPolicyClosed(
            'record_policy',
            then_=[CommunityOwners(), SystemProcess()],
            else_=[IfRestricted(
                'visibility',
                then_=[CommunityOwners()],
                else_=[AuthenticatedUser()]),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

from functools import wraps

from flask import g
from invenio_records_resources.services.errors import PermissionDeniedError

from invenio_communities.proxies import current_communities


def service():
    """Get the communities service."""
    return current_communities.service


def pass_community(f):
    """Decorate to retrieve the community record using the community service.
    """
    @wraps(f)
    def view(**kwargs):
        pid_value = kwargs.get('pid_value')
        community = service().read(
            id_=pid_value, identity=g.identity
        )
        kwargs['community'] = community
        return f(**kwargs)
    return view


def pass_community_logo(f):
    """Decorate a view to pass a community logo using the files service."""
    @wraps(f)
    def view(**kwargs):
        """."""
        try:
            pid_value = kwargs.get('pid_value')
            files = service().read_logo(
                id_=pid_value, identity=g.identity
            )
            kwargs['logo'] = files
        except FileNotFoundError:
            kwargs['logo'] = None

        return f(**kwargs)
    return view


def require_community_owner(f):
    """Decorate a view to require community owner for accessing the view."""
    @wraps(f)
    def view(**kwargs):
        """."""
        pid_value = kwargs.get('pid_value')
        community = service().read(
            id_=pid_value, identity=g.identity
        )
        if community.has_permissions_to(["update"])["can_update"]:
            return f(**kwargs)
        else:
            raise PermissionDeniedError()

    return view


# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

import re

from invenio_access.permissions import system_process
from invenio_pidstore.errors import PIDAlreadyExists
from invenio_records_resources.services.records.components import \
    ServiceComponent
from marshmallow.exceptions import ValidationError

from ...utils import on_membership_change

class PIDComponent(ServiceComponent):
    """Service component for Community PIDs."""

    @classmethod
    def _validate(cls, pid_value):
        """Checks the validity of the provided pid value."""
        blop = re.compile('^[-\w]+$')
        if not bool(blop.match(pid_value)):
            raise ValidationError(
                'The ID should contain only letters with numbers or dashes.',
                field_name='id',
            )

    def create(self, identity, record=None, data=None, **kwargs):
        """Create a Community PID from its metadata."""
        data['id'] = data['id'].lower()
        self._validate(data['id'])
        record['id'] = data['id']
        try:
            provider = record.__class__.pid.field._provider.create(record=record)
        except PIDAlreadyExists:
            raise ValidationError(
                'A community with this identifier already exists.',
                field_name='id',
            )
        setattr(record, 'pid', provider.pid)

    def update(self, identity, record=None, data=None, **kwargs):
        """Rename the Community PIDs value."""
        if 'id' in data and record.pid.pid_value != data['id']:
            raise ValidationError(
                'The ID should be modified through the renaming URL instead',
                field_name='id',
            )

    def rename(self, identity, record=None, data=None, **kwargs):
        """Rename the Community PIDs value."""
        data['id'] = data['id'].lower()
        if record.pid.pid_value == data['id']:
            raise ValidationError(
                'A new ID value is required for the renaming',
                field_name='id',
            )
        self._validate(data['id'])
        try:
            record.__class__.pid.field._provider.update(
                record.pid, data['id'])
        except PIDAlreadyExists:
            raise ValidationError(
                'A community with this identifier already exists.',
                field_name='id',
            )

# TODO: Move to Invenio-Records-Resources (and make reusable). Duplicated from
# Invenio-RDM-Records.
class AccessComponent(ServiceComponent):
    """Service component for access integration."""

    def _populate_access_and_validate(self, identity, data, record, **kwargs):
        """Populate and validate the record's access field."""
        if record is not None and "access" in data:
            # populate the record's access field with the data already
            # validated by marshmallow
            record.update({"access": data.get("access")})
            record.access.refresh_from_dict(record.get("access"))

        errors = record.access.errors
        if errors:
            # filter out duplicate error messages
            messages = list({str(e) for e in errors})
            raise ValidationError(messages, field_name="access")

    def _init_owners(self, identity, record, **kwargs):
        """If the record has no owners yet, add the current user."""
        # if the given identity is that of a user, we add the
        # corresponding user to the owners
        # (record.parent.access.owners) and commit the parent
        # otherwise, the parent's owners stays empty
        is_sys_id = system_process in identity.provides
        if not record.parent.access.owners and not is_sys_id:
            owner_dict = {"user": identity.id}
            record.parent.access.owners.add(owner_dict)
            record.parent.commit()

    def create(self, identity, data=None, record=None, **kwargs):
        """Add basic ownership fields to the record."""
        self._populate_access_and_validate(identity, data, record, **kwargs)
        self._init_owners(identity, record, **kwargs)

    def update_draft(self, identity, data=None, record=None, **kwargs):
        """Update handler."""
        self._populate_access_and_validate(identity, data, record, **kwargs)

    def publish(self, identity, draft=None, record=None, **kwargs):
        """Update draft metadata."""
        record.access = draft.access

    def edit(self, identity, draft=None, record=None, **kwargs):
        """Update draft metadata."""
        draft.access = record.access

    def new_version(self, identity, draft=None, record=None, **kwargs):
        """Update draft metadata."""
        draft.access = record.access


# TODO: Move to Invenio-Records-Resources (and make reusable). Duplicated from
# Invenio-RDM-Records.
class CommunityAccessComponent(AccessComponent):
    """Service component for access integration."""

    def _populate_access_and_validate(self, identity, data, record, **kwargs):
        """Populate and validate the community's access field."""
        if record is not None and "access" in data:
            # populate the record's access field with the data already
            # validated by marshmallow
            record.setdefault('access', {})
            record['access'].update(data.get("access", {}))
            record.access.refresh_from_dict(record.get("access"))

    def _init_owners(self, identity, record, **kwargs):
        """If the record has no owners yet, add the current user."""
        # if the given identity is that of a user, we add the
        # corresponding user to the owners (record.access.owned_by)
        is_sys_id = system_process in identity.provides
        if not record.access.owned_by and not is_sys_id:
            record.access.owned_by.add({"user": identity.id})
            on_membership_change(identity=identity)

    def create(self, identity, data=None, record=None, **kwargs):
        """Add basic ownership fields to the record."""
        self._populate_access_and_validate(identity, data, record, **kwargs)
        self._init_owners(identity, record, **kwargs)

    def update(self, identity, data=None, record=None, **kwargs):
        """Update handler."""
        self._populate_access_and_validate(identity, data, record, **kwargs)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Default configuration."""

COMMUNITIES_MEMBERSHIP_REQUESTS_CONFIRMLINK_EXPIRES_IN = 1000000

SUPPORT_EMAIL = 'test@email.org'

COMMUNITIES_RECORD_INDEX = 'records'

# TODO: Use vocabulary-based model
COMMUNITIES_DOMAINS = [
    {'text': 'Agricultural and Veterinary Sciences', 'value': 'agricultural_and_veterinary_sciences'},
    {'text': 'Biological Sciences', 'value': 'biological_sciences'},
    {'text': 'Built Environment and Design', 'value': 'built_environment_and_design'},
    {'text': 'Chemical Sciences', 'value': 'chemical_sciences'},
    {'text': 'Commerce, Management, Tourism and Services', 'value': 'commerce_management_tourism_and_services'},
    {'text': 'Earth Sciences', 'value': 'earth_sciences'},
    {'text': 'Environmental Sciences', 'value': 'environmental_sciences'},
    {'text': 'Economics', 'value': 'economics'},
    {'text': 'Education', 'value': 'education'},
    {'text': 'Engineering', 'value': 'engineering'},
    {'text': 'Technology', 'value': 'technology'},
    {'text': 'History and Archaeology', 'value': 'history_and_archaeology'},
    {'text': 'Information and Computing Sciences', 'value': 'information_and_computing_sciences'},
    {'text': 'Language, Communication and Culture', 'value': 'language_communication_and_culture'},
    {'text': 'Law and Legal Studies', 'value': 'law_and_legal_studies'},
    {'text': 'Mathematical Sciences', 'value': 'mathematical_sciences'},
    {'text': 'Medical and Health Sciences', 'value': 'medical_and_health_sciences'},
    {'text': 'Philosophy and Religious Studies', 'value': 'philosophy_and_religious_studies'},
    {'text': 'Physical Sciences', 'value': 'physical_sciences'},
    {'text': 'Psychology and Cognitive Sciences', 'value': 'psychology_and_cognitive_sciences'},
    {'text': 'Studies in Creative Arts and Writing', 'value': 'studies_in_creative_arts_and_writing'},
    {'text': 'Studies in Human Society', 'value': 'studies_in_human_society'},
]

COMMUNITIES_ROUTES = {
    'frontpage': '/communities',
    'search': '/communities/search',
    'new': '/communities/new',
    'details': '/communities/<pid_value>',
    'settings': '/communities/<pid_value>/settings',
    'settings_privileges': '/communities/<pid_value>/settings/privileges',
}
"""Communities ui endpoints."""

COMMUNITIES_ENABLED = True
"""Config to enable/disable communities blueprints."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

import itertools
import uuid

import pytest
from flask_security import login_user
from invenio_accounts.testutils import create_test_user, login_user_via_session
from invenio_app.factory import create_api

from invenio_communities.communities.records.api import Community


@pytest.fixture(scope='module')
def create_app(instance_path):
    """Application factory fixture."""
    return create_api


@pytest.fixture(scope='module')
def users(app):
    yield [create_test_user('user{}@inveniosoftware.org'.format(i)) for i in range(3)]


@pytest.fixture
def authenticated_user(app):
    """Authenticated user."""
    yield create_test_user('authed@inveniosoftware.org')


@pytest.fixture
def community_owner(app):
    """Record owner user."""
    yield create_test_user('community-owner@inveniosoftware.org')


@pytest.fixture(scope="function")
def minimal_community(community_owner):
    """Minimal community data as dict coming from the external world."""
    return {
        "id": "comm_id",
        "access": {
            "visibility": "public",
        },
        "metadata": {
            "title": "Title",
            "type": "topic"
        }
    }


@pytest.fixture(scope="function")
def full_community(community_owner):
    """Full community data as dict coming from the external world."""
    return  {
        "access": {
            "visibility": "public",
            "member_policy": "open",
            "record_policy": "open",
        },
        "id": "my_community_id",
        "metadata": {
            "title": "My Community",
            "description": "This is an example Community.",
            "type": "event",
            "curation_policy": "This is the kind of records we accept.",
            "page": "Information for my community.",
            "website": "https://inveniosoftware.org/",
            "organizations": [{
                    "name": "CERN",
            }]
        }
    }


@pytest.fixture()
def headers():
    """Default headers for making requests."""
    return {
        'content-type': 'application/json',
        'accept': 'application/json',
    }


@pytest.fixture()
def client_with_login(client, users):
    """Log in a user to the client."""
    user = users[0]
    login_user(user, remember=True)
    login_user_via_session(client, email=user.email)
    return client


@pytest.fixture(scope="function")
def create_many_records(app, client_with_login, minimal_community, headers):
    """Multiple community created and posted to test search functionality."""
    client = client_with_login
    community_types = ['organization', 'event', 'topic', 'project']
    N = 4
    for (type_,ind) in itertools.product(community_types, list(range(N))):
        minimal_community['id'] = f'comm_{type_}_{ind}'
        minimal_community['metadata']['type'] = type_
        client.post( '/communities', headers=headers, json=minimal_community)

    Community.index.refresh()

    # Return ids of first and last created communities
    return 'comm_organization_0', 'comm_project_3', N, N*len(community_types)
